import enum
import typing


class Level(enum.Enum):
    """
    An enum to represent severity of the :class:`DataSanityIssue`.
    """

    WARN = enum.auto()
    """
    Warning is an issue when something not entirely right. However, unlike :class:`Level.ERROR`,
    the analysis should complete albeit with sub-optimal results 😧.
    """

    ERROR = enum.auto()
    """
    Error is a serious issue in the input data and the downstream analysis may not complete or the analysis results
    may be malarkey 😱.
    """

    def __str__(self):
        return self.name


class Issue:
    """
    `Issue` summarizes an issue found in the input data.

    The issue has a :attr:`level`, a :attr:`message` with human-friendly description,
    and an optional :attr:`solution` for addressing the issue.
    """

    def __init__(
        self,
        level: Level,
        message: str,
        solution: typing.Optional[str] = None,
    ):
        self._level = level
        self._message = message
        self._solution = solution

    @property
    def level(self) -> Level:
        return self._level

    @property
    def message(self) -> str:
        return self._message

    @property
    def solution(self) -> typing.Optional[str]:
        return self._solution

    def __str__(self):
        return f"Issue(level={self._level}, message={self._message}, solution={self._solution})"

    def __repr__(self):
        return str(self)



"""
The module includes the :class:`Notepad` API
and the default implementation backed by a tree (not public).


Example
^^^^^^^

Create a new notepad with a top-level label `order`:

>>> from stairval.notepad import Notepad, create_notepad
>>> notepad = create_notepad('order')
>>> notepad.label
'order'

"""

from ._api import Notepad
from ._config import create_notepad

__all__ = [
    "Notepad",
    "create_notepad",
]



import pytest
import io
import os


from .simple import Person, Address, PersonAuditor, AddressAuditor


class TestAuditor:
    @pytest.fixture(scope="class")
    def auditor(self) -> PersonAuditor:
        return PersonAuditor(
            address_auditor=AddressAuditor(),
        )

    def test_ok_person(
        self,
        auditor: PersonAuditor,
    ):
        person = Person(
            name="Jack Black",
            age=43,
            address=Address(
                street="331 Farmington Ave",
                city="West Hartford",
                zip_code="06119",
                country="United States of America",
            ),
        )

        notepad = auditor.prepare_notepad("person")
        auditor.audit(person, notepad)

        buf = io.StringIO()
        notepad.summarize(file=buf)

        assert buf.getvalue() == "No errors or warnings were found" + os.linesep

    def test_few_errors(
        self,
        auditor: PersonAuditor,
    ):
        person = Person(
            name="Crooked Joe",
            age=-9,
            address=Address(
                street="",
                city="West Hartford",
                zip_code="-12345",
                country="United States of America",
            ),
        )

        notepad = auditor.prepare_notepad("person")
        auditor.audit(person, notepad)

        buf = io.StringIO()
        notepad.summarize(file=buf)

        expected = [
            "Showing errors and warnings",
            "  person",
            "  errors:",
            "  - `age` must not be negative",
            "    address",
            "    errors:",
            "    - `zip_code` must not be negative",
            "    warnings:",
            "    - `street` should not be empty",
        ]
        assert buf.getvalue() == (os.linesep.join(expected) + os.linesep)



import typing

from ._api import Notepad


class NotepadTree(Notepad):
    """
    `NotepadTree` implements :class:`Notepad` using a tree where each tree node corresponds to a (sub)section. The node
    can have `0..n` children.

    Each node has a :attr:`label`, a collection of issues, and children with subsections. For convenience, the node
    has :attr:`level` to correspond to the depth of the node within the tree (the level of the root node is `0`).

    The nodes can be accessed via :attr:`children` property or through convenience methods for tree traversal,
    either using the visitor pattern (:func:`visit`) or by iterating over the nodes via :func:`iterate_nodes`.
    In both cases, the traversal is done in the depth-first fashion.
    """

    # NOT PART OF THE PUBLIC API!

    def __init__(
        self,
        label: typing.Union[str, int],
        level: int,
    ):
        super().__init__(label, level)
        self._children: typing.MutableSequence["NotepadTree"] = []

    def add_subsection(self, label: typing.Union[str, int]) -> "NotepadTree":
        return NotepadTree._get_or_insert_node(self, label)

    def add_subsections(
        self,
        *labels: typing.Union[str, int],
    ) -> typing.Sequence["NotepadTree"]:
        nodes = []
        for label in labels:
            parent = self if len(nodes) == 0 else nodes[-1]
            node = NotepadTree._get_or_insert_node(parent, label)
            nodes.append(node)

        return nodes

    def get_subsections(self) -> typing.Sequence[Notepad]:
        return self._children

    @staticmethod
    def _get_or_insert_node(
        node: "NotepadTree",
        label: typing.Union[str, int],
    ) -> "NotepadTree":
        # Return the node if it already exists ...
        for child in node._children:
            if child.label == label:
                return child

        # ... or create a new node
        sub = NotepadTree(label, node._level + 1)
        node._children.append(sub)
        return sub

    def __str__(self):
        return f"NotepadTree(label={self._label}, level={self._level}, children={[ch.label for ch in self._children]})"

    def __repr__(self):
        return str(self)



import abc
import typing

from .notepad import Notepad, create_notepad

ITEM = typing.TypeVar("ITEM")
"""
The input for the :class:`Auditor`.
"""


class Auditor(typing.Generic[ITEM], metaclass=abc.ABCMeta):
    """
    `Auditor` checks the inputs for sanity issues and relates the issues with sanitized inputs
    as :class:`SanitationResults`.

    The auditor may sanitize the input as a matter of discretion and returns the input as `OUT`.
    """

    @staticmethod
    def prepare_notepad(label: str) -> Notepad:
        """
        Prepare a :class:`Notepad` for recording issues and errors.

        Args:
            label: a `str` with the top-level section label.

        Returns:
            Notepad: an instance of :class:`Notepad`.
        """
        return create_notepad(label)

    @abc.abstractmethod
    def audit(
        self,
        item: ITEM,
        notepad: Notepad,
    ):
        """
        Audit the `item` and record any issues into the `notepad`.
        """
        pass



"""
Stairval is a framework for validating hierarchical data structures.
"""

from importlib.metadata import version

from ._api import Level, Issue
from ._auditor import Auditor, ITEM

__version__ = version("stairval")

__all__ = [
    "Auditor",
    "Issue",
    "Level",
    "ITEM",
]



import os
import typing

import pytest

from stairval.notepad import Notepad, create_notepad


class TestNotepad:
    @pytest.fixture
    def notepad(self) -> Notepad:
        return create_notepad("badumtss")

    def test_add_subsection(
        self,
        notepad: Notepad,
    ):
        sub = notepad.add_subsection("subtss")

        assert sub.label == "subtss"
        assert sub.level == 1

    def test_subsection_labels_can_be_an_int(
        self,
        notepad: Notepad,
    ):
        assert not notepad.has_subsections()

        sub = notepad.add_subsection(321)

        assert sub.label == 321
        assert sub.level == 1

    def test_add_subsections(
        self,
        notepad: Notepad,
    ):
        assert not notepad.has_subsections()
        subs = notepad.add_subsections("foo", "bar", "baz")
        assert notepad.has_subsections()

        assert len(subs) == 3
        labels = [sub.label for sub in subs]
        assert labels == ["foo", "bar", "baz"]
        levels = [sub.level for sub in subs]
        assert levels == [1, 2, 3]
        all_labels = sorted(section.label for section in notepad.iter_sections())
        assert all_labels == ["badumtss", "bar", "baz", "foo"]

    def test_can_add_subsections_labeled_with_strs_and_ints(
        self,
        notepad: Notepad,
    ):
        assert not notepad.has_subsections()

        subs = notepad.add_subsections("foo", 493, "baz")
        assert notepad.has_subsections()

        assert len(subs) == 3
        assert [sub.label for sub in subs] == ["foo", 493, "baz"]

    def test_has_subsections(
        self,
        notepad: Notepad,
    ):
        assert not notepad.has_subsections()

        sub = notepad.add_subsection("subtss")
        assert not sub.has_subsections()

        assert notepad.has_subsections()

    def test_iter_sections_is_iterable(
        self,
        notepad: Notepad,
    ):
        assert isinstance(notepad.iter_sections(), typing.Iterable)

    def test_errors_is_iterable(
        self,
        notepad: Notepad,
    ):
        assert isinstance(notepad.errors(), typing.Iterable)

    def test_warnings_is_iterable(
        self,
        notepad: Notepad,
    ):
        assert isinstance(notepad.warnings(), typing.Iterable)

    def test_summary(
        self,
        notepad: Notepad,
    ):
        foo_pad = notepad.add_subsection("foo")
        foo_pad.add_error("A foo error")
        foo_pad.add_warning("A foo warning")

        bar_pad, pad_0, baz_pad = notepad.add_subsections("bar", 0, "baz")
        bar_pad.add_error("Bar error")
        pad_0.add_error("0 error")
        baz_pad.add_error("Baz error")

        summary = notepad.summary()
        lines = (
            "Showing errors and warnings",
            "  badumtss",
            "    foo",
            "    errors:",
            "    - A foo error",
            "    warnings:",
            "    - A foo warning",
            "    bar",
            "    errors:",
            "    - Bar error",
            "      0",
            "      errors:",
            "      - 0 error",
            "        baz",
            "        errors:",
            "        - Baz error",
            "",
        )
        assert summary == os.linesep.join(lines)

    def test_adding_the_same_subsection_twice_returns_the_same_subsection(
        self,
        notepad: Notepad,
    ):
        foo_one = notepad.add_subsection("foo")
        foo_two = notepad.add_subsection("foo")

        assert foo_one is foo_two

    def test_adding_the_same_subsections_twice_returns_the_same_subsections(
        self,
        notepad: Notepad,
    ):
        lefts = notepad.add_subsections("a", "b", 1, "c")
        rights = notepad.add_subsections("a", "b", 1, "c")

        assert all(left is right for left, right in zip(lefts, rights))



from ._api import Notepad
from ._tree import NotepadTree


def create_notepad(label: str) -> Notepad:
    """
    Create a default notepad.

    :param label: the label of the notepad
    """
    return NotepadTree(label=label, level=0)



import pytest

from ._tree import NotepadTree


class TestNotepadTree:
    @pytest.fixture(scope="class")
    def notepad(self) -> NotepadTree:
        return NotepadTree("label", 0)

    def test_tree_has_repr(
        self,
        notepad: NotepadTree,
    ):
        assert repr(notepad) == "NotepadTree(label=label, level=0, children=[])"

    def test_tree_has_str(
        self,
        notepad: NotepadTree,
    ):
        assert str(notepad) == "NotepadTree(label=label, level=0, children=[])"



import abc
import io
import os
import sys
import typing

from stairval import Issue, Level


class Notepad(metaclass=abc.ABCMeta):
    """
    Record issues encountered during parsing/validation of a hierarchical data structure.

    The issues can be organized in sections. `Notepad` keeps track of issues in one section
    and the subsections can be created by calling :func:`add_subsection`.
    The function returns an instance responsible for issues of a subsection.

    A collection of the issues from the current section are available via :attr:`issues` property
    and the convenience functions provide iterables over error and warnings.
    """

    def __init__(
        self,
        label: typing.Union[str, int],
        level: int,
    ):
        self._label = label
        self._level = level
        self._issues: typing.MutableSequence[Issue] = []

    @abc.abstractmethod
    def add_subsections(
        self,
        *labels: typing.Union[str, int],
    ) -> typing.Sequence["Notepad"]:
        """
        Add a sequence/chain of subsections.

        >>> from stairval.notepad import create_notepad
        >>> foo = create_notepad('foo')
        >>> subs = foo.add_subsections('bar', 0, 'baz')
        >>> len(subs)
        3

        Already existing subsections are not re-created:

        >>> subs2 = foo.add_subsections('bar', 0, 'bam')
        >>> subs[0] is subs2[0]
        True
        >>> subs[1] is subs2[1]
        True
        >>> subs[2] is subs2[2]
        False

        :param labels: a sequence of labels for the new notepad subsections.
        """
        pass

    @abc.abstractmethod
    def get_subsections(self) -> typing.Sequence["Notepad"]:
        """
        Get a sequence with subsections.

        Returns: a sequence of the subsection nodes.
        """
        ...

    def iter_sections(self) -> typing.Iterable["Notepad"]:
        """
        Iterate over nodes in the depth-first fashion.

        The iterable also includes the *current* node.

        Returns: a depth-first iterable over :class:`Notepad` nodes.
        """
        stack = [
            self,
        ]
        while stack:
            node = stack.pop()
            stack.extend(reversed(node.get_subsections()))  # type: ignore
            yield node

    def add_subsection(
        self,
        label: typing.Union[str, int],
    ) -> "Notepad":
        """
        Add a single labeled subsection.

        >>> from stairval.notepad import create_notepad
        >>> foo = create_notepad('foo')
        >>> bar = foo.add_subsection('bar')
        >>> bar.label
        'bar'

        If a subsection with the label already exists, then it is returned.

        >>> bar2 = foo.add_subsection('bar')
        >>> bar2 is bar
        True

        :param label: a label to use for the new subsection.
        """
        return self.add_subsections(label)[0]

    @property
    def label(self) -> typing.Union[str, int]:
        """
        Get the section label.
        """
        return self._label

    @property
    def level(self) -> int:
        """
        Get the level of the notepad node (distance from the root node, which has the level of `0`).
        """
        return self._level

    @property
    def issues(self) -> typing.Sequence[Issue]:
        """
        Get an iterable with the issues of the current section.
        """
        return self._issues

    def add_issue(self, level: Level, message: str, solution: typing.Optional[str] = None):
        """
        Add an issue with certain `level`, `message`, and an optional `solution`.
        """
        self._issues.append(Issue(level, message, solution))

    def add_error(self, message: str, solution: typing.Optional[str] = None):
        """
        A convenience function for adding an *error* with a `message` and an optional `solution`.
        """
        self.add_issue(Level.ERROR, message, solution)

    def errors(self) -> typing.Iterable[Issue]:
        """
        Iterate over the errors of the current section.
        """
        return filter(lambda dsi: dsi.level == Level.ERROR, self.issues)

    def error_count(self) -> int:
        """
        Returns:
            int: count of errors found in this section.
        """
        return sum(1 for _ in self.errors())

    def has_subsections(self) -> bool:
        """
        Returns:
            True: if the notepad has one or more subsections.
        """
        return len(self.get_subsections()) > 0

    def has_errors(self, include_subsections: bool = False) -> bool:
        """
        Returns:
            bool: `True` if one or more errors were found in the current section or its subsections.
        """
        if include_subsections:
            for node in self.iter_sections():
                for _ in node.errors():
                    return True
        else:
            for _ in self.errors():
                return True

        return False

    def add_warning(self, message: str, solution: typing.Optional[str] = None):
        """
        A convenience function for adding a *warning* with a `message` and an optional `solution`.
        """
        self.add_issue(Level.WARN, message, solution)

    def warnings(self) -> typing.Iterable[Issue]:
        """
        Iterate over the warnings of the current section.
        """
        return filter(lambda dsi: dsi.level == Level.WARN, self.issues)

    def has_warnings(self, include_subsections: bool = False) -> bool:
        """
        Returns:
            bool: `True` if one or more warnings were found in the current section or its subsections.
        """
        if include_subsections:
            for node in self.iter_sections():
                for _ in node.warnings():
                    return True
        else:
            for _ in self.warnings():
                return True

        return False

    def warning_count(self) -> int:
        """
        Returns:
            int: count of warnings found in this section.
        """
        return sum(1 for _ in self.warnings())

    def has_errors_or_warnings(self, include_subsections: bool = False) -> bool:
        """
        Returns:
            bool: `True` if one or more errors or warnings were found in the current section or its subsections.
        """
        if include_subsections:
            for node in self.iter_sections():
                for _ in node.warnings():
                    return True
                for _ in node.errors():
                    return True
        else:
            for _ in self.warnings():
                return True
            for _ in self.errors():
                return True

        return False

    def visit(
        self,
        visitor: typing.Callable[
            [
                "Notepad",
            ],
            None,
        ],
    ):
        """
        Performs a depth-first search on the notepad nodes and calls `visitor` with all nodes.
        Args:
            visitor: a callable that takes the current notepad node as the only argument.
        """
        for node in self.iter_sections():
            visitor(node)

    def summarize(
        self,
        file: typing.TextIO = sys.stdout,
        indent: int = 2,
    ):
        """
        Summarize the notepad into `file` (STDOUT by default).

        :param file: a TextIO-like object to write the summary into (STDOUT by default).
        :param indent: the number of spaces to delimit the notepad subsections (default: `2`).
        """
        assert isinstance(indent, int) and indent >= 0

        n_errors = sum(node.error_count() for node in self.iter_sections())
        n_warnings = sum(node.warning_count() for node in self.iter_sections())
        if n_errors > 0 or n_warnings > 0:
            file.write("Showing errors and warnings")
            file.write(os.linesep)

            for node in self.iter_sections():
                if node.has_errors_or_warnings(include_subsections=True):
                    # We must report the node label even if there are no issues with the node.
                    l_pad = " " * ((node.level + 1) * indent)
                    file.write(l_pad + str(node.label))
                    file.write(os.linesep)

                    if node.has_errors():
                        file.write(l_pad + "errors:")
                        file.write(os.linesep)
                        for error in node.errors():
                            file.write(l_pad + "- " + error.message + (f"· {error.solution}" if error.solution else ""))
                            file.write(os.linesep)
                    if node.has_warnings():
                        file.write(l_pad + "warnings:")
                        file.write(os.linesep)
                        for warning in node.warnings():
                            file.write(
                                l_pad + "- " + warning.message + (f"· {warning.solution}" if warning.solution else "")
                            )
                            file.write(os.linesep)
        else:
            file.write("No errors or warnings were found")
            file.write(os.linesep)

    def summary(
        self,
        indent: int = 2,
    ) -> str:
        """
        Summarize the notepad into a `str`.

        :param indent: the number of spaces to delimit the notepad subsections (default: `2`).
        :return: The notepad summary.
        """
        buf = io.StringIO()
        self.summarize(file=buf, indent=indent)
        return buf.getvalue()
